package com.imcode.forum.sql;

import com.imcode.db.Database;
import com.imcode.db.DatabaseCommand;
import com.imcode.db.DatabaseConnection;
import com.imcode.db.DatabaseException;
import com.imcode.db.DatasourceDatabase;
import com.imcode.db.commands.InsertIntoTableDatabaseCommand;
import com.imcode.db.commands.SqlQueryDatabaseCommand;
import com.imcode.db.commands.SqlUpdateDatabaseCommand;
import com.imcode.db.commands.TransactionDatabaseCommand;
import com.imcode.db.commands.UpdateTableWhereColumnEqualsDatabaseCommand;
import com.imcode.db.exceptions.IntegrityConstraintViolationException;
import com.imcode.db.handlers.ObjectFromFirstRowResultSetHandler;
import com.imcode.forum.Forum;
import com.imcode.forum.ForumPost;
import com.imcode.forum.ForumRepository;
import com.imcode.forum.ForumThread;
import com.imcode.forum.Id;
import org.apache.commons.lang.ClassUtils;

import javax.sql.DataSource;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class SqlRepository implements ForumRepository {

    private static final String SQL_STATE__MISSING_TABLE = "42S02";

    private Database database;
    private static final String TABLE_POSTS = "forum_posts";
    private static final String TABLE_THREADS = "forum_threads";
    private static final String TABLE_FORUMS = "forum_forums";

    public String toString() {
        return ClassUtils.getShortClassName(getClass())+"("+database+")" ;
    }

    public SqlRepository(Database database) {
        this.database = database;
    }

    public SqlRepository(DataSource dataSource) {
        this(new DatasourceDatabase(dataSource)) ;
    }

    public void dropTables() {
        database.executeCommand(new DatabaseCommand() {
            public Object executeOn(DatabaseConnection connection) throws DatabaseException {
                try {
                    Set<String> tableNames = getTableNames(connection, "forum_%");
                    dropTable(connection, TABLE_POSTS, tableNames) ;
                    dropTable(connection, TABLE_THREADS, tableNames) ;
                    dropTable(connection, TABLE_FORUMS, tableNames) ;
                    return null ;
                } catch ( SQLException e ) {
                    throw DatabaseException.fromSQLException(null, e) ;
                }
            }

            private void dropTable(DatabaseConnection connection, String tableName, Set<String> tableNames) {
                if (tableNames.contains(tableName)) {
                    connection.executeUpdate("DROP TABLE "+tableName, null) ;
                }
            }
        }) ;
    }

    public void createTables() {
        final String databaseProductName = getDatabaseProductName();
        
        database.executeCommand(new DatabaseCommand() {
            public Object executeOn(DatabaseConnection connection) throws DatabaseException {
                try {
                    ForumDdl forumDdl = null ;
                    if ( "mysql".equalsIgnoreCase(databaseProductName) ) {
                        forumDdl = new MysqlForumDdl();
                    } else if ("microsoft sql server".equalsIgnoreCase(databaseProductName)) {
                        forumDdl = new MssqlForumDdl();
                    } else {
                        throw new DatabaseException("Unsupported database: "+databaseProductName, null);
                    }
                    Set<String> tableNames = getTableNames(connection, "forum_%");
                    createTable(connection, TABLE_FORUMS, forumDdl.getForumsDdl(), tableNames);
                    createTable(connection, TABLE_THREADS, forumDdl.getThreadsDdl(), tableNames);
                    createTable(connection, TABLE_POSTS, forumDdl.getPostsDdl(), tableNames);
                } catch( SQLException e ) {
                    throw DatabaseException.fromSQLException(null, e) ;
                }
                return null ;
            }
        }) ;
    }

    private String getDatabaseProductName() {
        return (String) database.executeCommand(new DatabaseCommand() {
            public Object executeOn(DatabaseConnection connection) throws DatabaseException {
                try {
                    DatabaseMetaData metaData = connection.getConnection().getMetaData();
                    return metaData.getDatabaseProductName() ;
                } catch ( SQLException e ) {
                    throw DatabaseException.fromSQLException(null, e);
                }
            }
        });
    }

    private void createTable(DatabaseConnection connection, String tableName, String tableDdl, Set<String> tableNames) {
        if (!tableNames.contains(tableName)) {
            connection.executeUpdate(tableDdl, null) ;
        }
    }

    private Set<String> getTableNames(DatabaseConnection connection, String tableNamePattern) throws SQLException {
        DatabaseMetaData metaData = connection.getConnection().getMetaData();
        ResultSet tablesResultSet = metaData.getTables(null, null, tableNamePattern, null);
        Set<String> tableNames = new HashSet<String>() ;
        while (tablesResultSet.next()) {
            String tableName = tablesResultSet.getString("TABLE_NAME") ;
            tableNames.add(tableName) ;
        }
        return tableNames;
    }

    public Forum getForum(Id id) {
        return (Forum) database.executeCommand(
                new SqlQueryDatabaseCommand(
                        "SELECT forum_id from forum_forums WHERE forum_id = ?",
                        new Object[] { id.getValue() },
                        new ObjectFromFirstRowResultSetHandler(new SqlForumFactory(this)))) ;
    }

    ForumThread createThread(final Id forumId, final ForumPost post) {
        Id threadId = (Id) database.executeCommand(new TransactionDatabaseCommand() {
            public Object executeInTransaction(DatabaseConnection connection) throws DatabaseException {
                Id threadId = new SqlId((Number) new InsertIntoTableDatabaseCommand(TABLE_THREADS, new Object[][]{
                        { "forum_id", forumId.getValue() }
                }).executeOn(connection) ) ;

                createPost(connection, threadId, null, post);
                return threadId ;
            }
        }) ;
        return new SqlThread(threadId, this) ;
    }

    List<ForumThread> getForumThreads(Id forumId) {
        return (List<ForumThread>) database.executeCommand(
                new SqlQueryDatabaseCommand(
                        "SELECT ft.thread_id\n"
                        + "FROM forum_threads ft\n"
                        + "JOIN (SELECT thread_id, MAX(datetime) datetime FROM forum_posts GROUP BY thread_id) np\n"
                        + "ON ft.thread_id = np.thread_id\n"
                        + "WHERE ft.forum_id = ?\n"
                        + "ORDER BY np.datetime DESC\n",
                        new Object[] { forumId.getValue() },
                        new CollectionResultSetHandler(new SqlThreadFactory(this), new ArrayList()))) ;
    }


    public List<Forum> getForums() {
        return (List<Forum>) database.executeCommand(
                new SqlQueryDatabaseCommand(
                        "SELECT forum_id FROM forum_forums",
                        null,
                        new CollectionResultSetHandler(new SqlForumFactory(this), new ArrayList()))) ;
    }

    public Forum createForum(final Id forumId) {
        try {
            tryCreateForum(forumId);
        } catch( DatabaseException dbe ) {
            for ( Throwable t = dbe ; null != t; t = t.getCause() ) {
                if (t instanceof IntegrityConstraintViolationException) {
                    return getForum(forumId) ;
                } else if (t instanceof SQLException && SQL_STATE__MISSING_TABLE.equals(((SQLException)t ).getSQLState())) {
                    createTables();
                    tryCreateForum(forumId);
                    return new SqlForum(this, forumId);
                }
            }
            throw dbe ;
        }
        return new SqlForum(this, forumId);
    }

    private void tryCreateForum(Id forumId) {
        database.executeCommand(new InsertIntoTableDatabaseCommand(TABLE_FORUMS, new Object[][] {
            { "forum_id", forumId.getValue() }
        })) ;
    }

    public Id createId(Object value) {
        if (value instanceof String) {
            return new SqlId(Long.valueOf((String) value));
        }
        return new SqlId((Number)value);
    }

    ForumThread getThread(Id threadId) {
        return (ForumThread) database.executeCommand(
                new SqlQueryDatabaseCommand(
                        "SELECT forum_threads.thread_id FROM forum_threads WHERE forum_threads.thread_id = ?",
                        new Object[] { threadId.getValue()},
                        new ObjectFromFirstRowResultSetHandler(new SqlThreadFactory(this))) ) ;
    }

    private Id createPost(DatabaseConnection connection, Id threadId, Id parentPostId, ForumPost post) {
        return new SqlId((Number) new InsertIntoTableDatabaseCommand(TABLE_POSTS, new Object[][] {
                { "thread_id", threadId.getValue() },
                { "parent_post_id", null == parentPostId ? null : parentPostId.getValue() },
                { "datetime", new Timestamp(post.getDateTime().getTime()) },
                { "userdata", post.getUser() },
                { "title", post.getTitle() },
                { "body", post.getBody() },
                { "hidden", new Boolean(post.isHidden()) },
        }).executeOn(connection) ) ;
    }

    void incrementPostViewCount(Id postId) {
        database.executeCommand(
                new SqlUpdateDatabaseCommand(
                        "UPDATE forum_posts SET view_count = view_count + 1 WHERE post_id = ?",
                        new Object[] { postId.getValue() })) ;
    }

    Id createPostReply(final Id threadId, final Id postId, final ForumPost forumPost) {
       return (Id) database.executeCommand(new DatabaseCommand() {
           public Object executeOn(DatabaseConnection connection) throws DatabaseException {
               return createPost(connection, threadId, postId, forumPost );
           }
       }) ;
    }

    SqlPostsSet getThreadPosts(SqlThread thread) {
        return (SqlPostsSet) database.executeCommand(
                new SqlQueryDatabaseCommand("SELECT "+ SqlPostFactory.POST_COLUMNS+" FROM " + TABLE_POSTS + " WHERE thread_id = ? ORDER BY datetime, post_id",
                                            new Object[] { thread.getId().getValue() },
                                            new CollectionResultSetHandler(new SqlPostFactory(thread), new SqlPostsSet()))) ;
    }

    void hidePost(Id sqlThread, Id post, boolean hidden) {
        database.executeCommand(new UpdateTableWhereColumnEqualsDatabaseCommand(TABLE_POSTS, "post_id", post.getValue(), new Object[][] {
                { "hidden", new Boolean(hidden) }
        })) ;
    }
}
